#!/usr/bin/python

# $Id: tfd128.py 36 2007-09-26 17:01:26Z ae $
#
# Copyright (C) 2007 Andreas Engel
# Copyright (C) 2007 Pascal Bihler (Mac OS X adaption)
#
# See README for installation instructions.
#
# TODO:
# - localization
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import time, sys, os

class Tfd128:

    MODE_TEMP_MASK = 0x02
    MODE_HUM_MASK  = 0x01

    class Busy(Exception)  : pass
    class Error(Exception) : pass

    def __init__(self, device = ''):
        if device == '':
            for name in [
                '/dev/tfd128'			# Linux, preferred
                '/dev/tty.usbserial-3B1',	# Mac OS X
                '/dev/ttyUSB0',			# Linux, fallback
                ]:
                if os.path.exists(name):
                    device = name
                    break
            else:
                 raise self.Error("TFD 128 device not found")

        self._device = device
        self._count  = 0
        self._mode   = 0
        self._debug  = False

    def __xfer(self, cmd, data=[]):

        """transfers the given cmd and data values to and returns the answer
        from the device"""

        import serial

        # some constants for better readability
        STX = 0x02
        ETX = 0x03
        ENQ = 0x05
        ACK = 0x06
        NAK = 0x15

        ser = serial.Serial(self._device,
                            baudrate=38400,
                            timeout=5,
                            parity=serial.PARITY_EVEN,
                            stopbits=serial.STOPBITS_ONE)

        # send command sequence
        ser.write("%c" % STX)
        ser.write(cmd)
        for c in data:
            if c == STX or c == ETX or c == ENQ:
                ser.write("%c" % ENQ)
                c += 0x80
            ser.write("%c" % c)
        ser.write("%c" % ETX)

        # Answer sequence must start with STX.
        r = ord(ser.read(1))
        if r != STX:
            raise self.Error("internal error: expected STX, got %s" % r)

        # Then the command itself.
        r = ord(ser.read(1))
        if r != ord(cmd):
            raise self.Error("internal error: expected %s, got %s" % (cmd, r))

        result = []
        while r != ETX:
            r = ord(ser.read(1))
            if r == ENQ:
                r = ord(ser.read(1))
                result.append(r - 0x80)
            elif r != ETX:
                result.append(r)

        ser.close()

        if self._debug: print result

        if result[0] == NAK:
            raise self.Busy

        if len(data) > 0 and result[0] != ACK:
            raise self.Error("internal error: expected ACK got %s" % result[0])

        return result

    def get_status(self):
        """Use this method to determine whether the TFD128 is idle or not.
        Returns the either the string IDLE or BUSY"""
        try:
            result = self.__xfer('V')
            return 'IDLE'
        except self.Busy:
            return 'BUSY'

    def get_count(self):
        """Returns the number of data points"""
        result = self.__xfer('A')
        return result[0] + (result[1] << 8)

    def start(self, interval, mode):
        result = self.__xfer('S', self.__make_date() + [mode, interval])
        return 'OK'

    def stop(self):
        result = self.__xfer('E', self.__make_date())
        return 'OK'

    def __make_date(self):
        from datetime import datetime
        now = datetime.now()
        return [ now.year & 0xFF, now.year >> 8, now.month - 1, now.day,
                 now.hour, now.minute, now.second ]

    def __make_temp(self, temp):
        (lo, hi) = temp
        temp = lo + (hi << 8)
        # negative values need sign extension
        temp = temp - ((temp & 0x8000) << 1)
        # values are stored in 1/10 degrees celsius
        temp = temp / 10.0
        return temp

    def __get_block(self, cmd):
        if not self._count: return []
        result = self.__xfer(cmd)
        r = []
        # in most cases the device returns more values than
        # available; that's why we need to count down here
        if self._mode & self.MODE_HUM_MASK:
            for i in xrange(0, len(result)-1, 3):
                temp = self.__make_temp(result[i:i+2])
                r.append((temp, result[i+2]))
                self._count -= 1
                if not self._count: break
        else:
            for i in xrange(0, len(result)-1, 2):
                temp = self.__make_temp(result[i:i+2])
                r.append(temp)
                self._count -= 1
                if not self._count: break
        return r

    # we split getting the values into two methods; this can later
    # be used in the GUI application for providing a status feedback
    # when retrieving the data values
    def get_values(self):
        if self._start:
            self._start = False
            return self.__get_block('R')
        else:
            return self.__get_block('N')

    def get_values_init(self):
        start, stop, mode, interval = self.get_info()
        self._mode  = mode
        self._count = self.get_count()
        self._start = True
        if stop == 0:
            # no stop time (battery failure or full)
            stop = start + (self._count - 1) * interval * 60
        else:
            # fix stop time: as the device only stores measurement points in
            # fixed intervals, the last stored measurement point may be 1 or
            # 5 minutes (minus 1s) before the stored end time; we therefore
            # need to round down the stop time
            from math import floor
            stop = start + floor((stop - start)/60/interval) * 60 * interval
        if self._debug:
            print "Real start: %s" % time.asctime(time.localtime(start))
            print "Real stop:  %s" % time.asctime(time.localtime(stop))
        return (mode, start, stop, self._count)

    def get_info(self):
        result = self.__xfer('Z')
        start  = self.__parse_date(result[0:6+1])
        if len(result) > 9:
            stop = self.__parse_date(result[9:15+1])
        else:
            # no stop time recorded
            stop = 0
        mode     = result[7]
        interval = result[8]
        return (start, stop, mode, interval)

    def __parse_date(self, date):
        date = [date[0]+(date[1]<<8), date[2]+1] + date[3:6+1] + [0, 1, -1]
        return time.mktime(date)

    def get_version(self):
        result = self.__xfer('V')
        return result[0] + (result[1] << 8)

    def debug_on(self):
        self._debug = True

if __name__ == '__main__':
    import optparse
    p = optparse.OptionParser()
    p.add_option('--debug', '-d', action='store_true',
                 help='for debugging only')
    p.add_option('--start', '-S', action='store_true',
                 help='start measurement')
    p.add_option('--stop', '-E', action='store_true',
                 help='stop measurement')
    p.add_option('--interval', '-i', type='choice', action='store',
                 choices=('1', '5'),
                 help='measurement interval, defaults to 5min')
    p.add_option('--mode', '-m', type='choice', action='store',
                 choices=('t', 'tf', 'ft', 'th', 'ht'),
                 help='measurement mode, defaults to temp+humidity')
    p.add_option('--status', '-s', action='store_true',
                 help='print current status')
    p.add_option('--dump-version', '-v', action='store_true',
                 help='print tfd128 version')
    p.add_option('--dump-count', '-a', action='store_true',
                 help='print number of collected data points')
    p.add_option('--dump-values', '-r', action='store_true',
                 help='print data points to ./tfd128-<date>.csv as CSV')
    p.add_option('--output', '-o', action='store',
                 help='output to given filename; "-" = stdout')
    p.add_option('--no-progress', '-p', action='store_true',
                 help='disable progress bar when dumping values')
    p.add_option('--no-heading', '-n', action='store_true',
                 help="suppress heading line when dumping values")
    p.add_option('--dump-info', '-z', action='store_true',
                 help='print data record info')

    p.set_defaults(mode='tf', interval='5')

    opts, args = p.parse_args()

    if len(args) != 0:
        p.error("too many arguments")

    try:
        logger = Tfd128()

        if opts.debug: logger.debug_on()

        if opts.start:
            mode = 0
            if opts.mode.find('t') >= 0: mode |= logger.MODE_TEMP_MASK
            if opts.mode.find('f') >= 0: mode |= logger.MODE_HUM_MASK
            if opts.mode.find('h') >= 0: mode |= logger.MODE_HUM_MASK
            print "Status = %s" % logger.start(int(opts.interval), mode)
            sys.exit(0)

        if opts.stop:
            print "%s" % logger.stop()
            sys.exit(0)

        if opts.dump_version:
            print "%s" % logger.get_version()
            sys.exit(0)

        if opts.dump_count:
            print "%s" % logger.get_count()
            sys.exit(0)

        if opts.dump_info:
            start, stop, mode, interval = logger.get_info()
            start = time.asctime(time.localtime(start))
            if stop == 0:
                stop = "<no time recorded>"
            else:
                stop  = time.asctime(time.localtime(stop))
            if mode & logger.MODE_HUM_MASK:
                mode = 'temperature+humidity'
            else:
                mode = 'temperature'
            print "Start : %s" % start
            print "Stop  : %s" % stop
            print "Mode  : %s" % mode
            print "Intvl : %d min" % interval
            sys.exit(0)

        if opts.dump_values:
            if not opts.no_progress: import progress
            mode, start, stop, count = logger.get_values_init()
            if opts.output == '-':
                output = sys.stdout
            else:
                if opts.output == None:
                    filename = time.strftime("tfd128-%Y%m%d.csv",
                                             time.localtime(stop))
                    print "Data will be written to file '%s'" % filename
                else:
                    filename = opts.output
                output = open(filename, 'w')
            if count == 1:
                delta = 0
            else:
                delta = (stop - start) / (count - 1)
            if not opts.no_progress:
                progress = progress.ProgressDisplay(count, sys.stderr)
            values = logger.get_values()
            if not opts.no_progress: progress += len(values)
            actual = 0
            if not opts.no_heading:
                output.write("Nummer;Datum;Temperatur")
                if mode & logger.MODE_HUM_MASK:
                    output.write(";Luftfeuchtigkeit")
                output.write("\n")
            while values:
                for v in values:
                    t = start + actual * delta
                    t = time.strftime("%d.%m.%Y %H:%M:%S",time.localtime(t))
                    if mode & logger.MODE_HUM_MASK:
                        v = "%4.1f;%d" % (v[0], v[1])
                    else:
                        v = "%4.1f" % v
                    output.write("%d;%s;%s\n" % (actual, t, v))
                    actual += 1
                values = logger.get_values()
                if not opts.no_progress: progress += len(values)
            sys.exit(0)

        if opts.status:
            status = logger.get_status()
            print status
            if status == 'IDLE':
                sys.exit(0)
            else:
                sys.exit(1)

    except Tfd128.Error, exception:
        print exception
        sys.exit(1)

    except Tfd128.Busy:
        print 'Busy'
        sys.exit(1)

    p.print_help()
